@@ -1,7 +1,7 @@
 /*
  * ndpi_api.h
  *
- * Copyright (C) 2011-20 - ntop.org
+ * Copyright (C) 2011-24 - ntop.org
  *
  * This file is part of nDPI, an open source deep packet inspection
  * library based on the OpenDPI and PACE technology by ipoque GmbH
@@ -115,17 +115,17 @@ extern "C" {
   u_int32_t ndpi_get_tot_allocated_memory(void);
 
   /**
-   * Search the first occurrence of substring -find- in -s-
-   * The search is limited to the first -slen- characters of the string
+   * Finds the first occurrence of the substring 'needle' in the string 'haystack'.
    *
-   * @par    s     = string to parse
-   * @par    find  = string to match with -s-
-   * @par    slen  = max length to match between -s- and -find-
-   * @return a pointer to the beginning of the located substring;
-   *         NULL if the substring is not found
+   * This function is similar to the standard `strstr()` function, but it has an additional parameter `len` that
+   * specifies the maximum length of the search.
    *
+   * @param haystack The string to search in.
+   * @param needle The substring to search for.
+   * @param len The maximum length of the search.
+   * @return Pointer to the first occurrence of 'needle' in 'haystack', or NULL if no match is found.
    */
-  char* ndpi_strnstr(const char *s, const char *find, size_t slen);
+  char *ndpi_strnstr(const char *haystack, const char *needle, size_t len);
 
   /**
    * Same as ndpi_strnstr but case insensitive
@@ -165,6 +165,9 @@ extern "C" {
   u_int16_t ndpi_network_port_ptree_match(struct ndpi_detection_module_struct *ndpi_struct,
 					  struct in_addr *pin /* network byte order */,
 					  u_int16_t port /* network byte order */);
+  u_int16_t ndpi_network_port_ptree6_match(struct ndpi_detection_module_struct *ndpi_struct,
+					   struct in6_addr *pin,
+					   u_int16_t port /* network byte order */);
 
   /**
    * Creates a protocol match that does not contain any hostnames.
@@ -198,6 +201,22 @@ extern "C" {
   void ndpi_init_protocol_match(struct ndpi_detection_module_struct *ndpi_mod,
                                 ndpi_protocol_match const * const match);
 
+  /**
+   * Returns a new initialized global context.
+   *
+   * @return  the initialized global context
+   *
+   */
+  struct ndpi_global_context *ndpi_global_init(void);
+
+  /**
+   * Deinit a properly initialized global context.
+   *
+   * @par g_ctx = global context to free/deinit
+   *
+   */
+  void ndpi_global_deinit(struct ndpi_global_context *g_ctx);
+
   /**
    * Returns a new initialized detection module
    * Note that before you can use it you can still load
@@ -208,19 +227,21 @@ extern "C" {
    * indipendent detection contexts) but all these calls MUST NOT run
    * in parallel
    *
-   * @par prefs = load preferences
+   * @g_ctx = global context associated to the new detection module; NULL if no global context is needed
    * @return  the initialized detection module
    *
    */
-  struct ndpi_detection_module_struct *ndpi_init_detection_module(ndpi_init_prefs prefs);
+  struct ndpi_detection_module_struct *ndpi_init_detection_module(struct ndpi_global_context *g_ctx);
 
   /**
    * Completes the initialization (2nd step)
    *
    * @par ndpi_str = the struct created for the protocol detection
    *
+   * @return 0 on success
+   *
    */
-  void ndpi_finalize_initialization(struct ndpi_detection_module_struct *ndpi_str);
+  int ndpi_finalize_initialization(struct ndpi_detection_module_struct *ndpi_str);
 
   /**
    * Frees the dynamic memory allocated members in the specified flow
@@ -238,18 +259,6 @@ extern "C" {
    */
   void ndpi_free_flow(struct ndpi_flow_struct *flow);
 
-  /**
-   * Enables cache support.
-   * In nDPI is used for some protocol (i.e. Skype)
-   *
-   * @par ndpi_mod  = the struct created for the protocol detection
-   * @par host      = string for the host name
-   * @par port      = unsigned int for the port number
-   *
-   */
-  void ndpi_enable_cache(struct ndpi_detection_module_struct *ndpi_mod,
-			 char* host, u_int port);
-
   /**
    * Destroys the detection module
    *
@@ -298,14 +307,12 @@ extern "C" {
    *
    * @par    ndpi_struct  = the detection module
    * @par    flow         = the flow given for the detection module
-   * @par    enable_guess = guess protocol if unknown
    * @par    protocol_was_guessed = 1 if the protocol was guesses (requires enable_guess = 1), 0 otherwise
    * @return the detected protocol even if the flow is not completed;
    *
    */
   ndpi_protocol ndpi_detection_giveup(struct ndpi_detection_module_struct *ndpi_struct,
 				      struct ndpi_flow_struct *flow,
-				      u_int8_t enable_guess,
 				      u_int8_t *protocol_was_guessed);
 
   /**
@@ -352,64 +359,42 @@ extern "C" {
    * Get the main protocol of the passed flows for the detected module
    *
    *
-   * @par    ndpi_struct  = the detection module
    * @par    flow         = the flow given for the detection module
    * @return the ID of the master protocol detected
    *
    */
-  u_int16_t ndpi_get_flow_masterprotocol(struct ndpi_detection_module_struct *ndpi_struct,
-					 struct ndpi_flow_struct *flow);
+  u_int16_t ndpi_get_flow_masterprotocol(struct ndpi_flow_struct *flow);
 
   /**
    * Get the app protocol of the passed flows for the detected module
    *
    *
-   * @par    ndpi_struct  = the detection module
    * @par    flow         = the flow given for the detection module
    * @return the ID of the app protocol detected
    *
    */
-  u_int16_t ndpi_get_flow_appprotocol(struct ndpi_detection_module_struct *ndpi_str, struct ndpi_flow_struct *flow);
+  u_int16_t ndpi_get_flow_appprotocol(struct ndpi_flow_struct *flow);
 
   /**
    * Get the category of the passed flows for the detected module
    *
    *
-   * @par    ndpi_struct  = the detection module
    * @par    flow         = the flow given for the detection module
    * @return the ID of the category
    *
    */
-  ndpi_protocol_category_t ndpi_get_flow_category(struct ndpi_detection_module_struct *ndpi_str, struct ndpi_flow_struct *flow);
+  ndpi_protocol_category_t ndpi_get_flow_category(struct ndpi_flow_struct *flow);
 
   /**
    * Get the ndpi protocol data of the passed flows for the detected module
    *
    *
-   * @par    ndpi_struct  = the detection module
    * @par    flow         = the flow given for the detection module
    * @par    ndpi_proto   = the output struct where to store the requested information
    *
    */
-  void ndpi_get_flow_ndpi_proto(struct ndpi_detection_module_struct *ndpi_str, struct ndpi_flow_struct *flow,
-                    struct ndpi_proto * ndpi_proto);
-
-  /**
-   * API call that is called internally by ndpi_detection_process_packet or by apps
-   * that want to avoid calling ndpi_detection_process_packet as they have already
-   * parsed the packet and thus want to avoid this.
-   *
-   *
-   * @par    ndpi_struct              = the detection module
-   * @par    flow                     = the flow given for the detection module
-   * @par    ndpi_selection_bitmask   = the protocol selected bitmask
-   * @return number of protocol dissector that have been tried (0 = no more dissectors)
-   *
-   */
-  u_int32_t ndpi_check_flow_func(struct ndpi_detection_module_struct *ndpi_struct,
-				 struct ndpi_flow_struct *flow,
-				 NDPI_SELECTION_BITMASK_PROTOCOL_SIZE *ndpi_selection_packet);
-
+  void ndpi_get_flow_ndpi_proto(struct ndpi_flow_struct *flow,
+				struct ndpi_proto * ndpi_proto);
 
   /**
    * Query the pointer to the layer 4 packet
@@ -424,7 +409,8 @@ extern "C" {
    else != 0
    *
    */
-  u_int8_t ndpi_detection_get_l4(const u_int8_t *l3, u_int16_t l3_len, const u_int8_t **l4_return, u_int16_t *l4_len_return,
+  u_int8_t ndpi_detection_get_l4(const u_int8_t *l3, u_int16_t l3_len,
+				 const u_int8_t **l4_return, u_int16_t *l4_len_return,
 				 u_int8_t *l4_protocol_return, u_int32_t flags);
 
   /**
@@ -439,10 +425,8 @@ extern "C" {
    *
    */
   ndpi_protocol ndpi_find_port_based_protocol(struct ndpi_detection_module_struct *ndpi_struct/* , u_int8_t proto */,
-					      u_int32_t shost,
-					      u_int16_t sport,
-					      u_int32_t dhost,
-					      u_int16_t dport);
+					      u_int32_t shost, u_int16_t sport,
+					      u_int32_t dhost, u_int16_t dport);
   /**
    * Search and return the protocol guessed that is undetected
    *
@@ -472,10 +456,8 @@ extern "C" {
   ndpi_protocol ndpi_guess_undetected_protocol_v4(struct ndpi_detection_module_struct *ndpi_struct,
 						  struct ndpi_flow_struct *flow,
 						  u_int8_t proto,
-						  u_int32_t shost,
-						  u_int16_t sport,
-						  u_int32_t dhost,
-						  u_int16_t dport);
+						  u_int32_t shost, u_int16_t sport,
+						  u_int32_t dhost, u_int16_t dport);
   /**
    * Check if the string passed match with a protocol
    *
@@ -561,15 +543,13 @@ extern "C" {
    * Same as ndpi_protocol2name() with the difference that the numeric protocol
    * name is returned
    *
-   * @par     ndpi_mod      = the detection module
    * @par     proto         = the struct ndpi_protocol contain the protocols name
    * @par     buf           = the buffer to write the name of the protocols
    * @par     buf_len       = the length of the buffer
    * @return  the buffer contains the master_protocol and protocol name
    *
    */
-  char* ndpi_protocol2id(struct ndpi_detection_module_struct *ndpi_mod,
-			 ndpi_protocol proto, char *buf, u_int buf_len);
+  char* ndpi_protocol2id(ndpi_protocol proto, char *buf, u_int buf_len);
 
   /**
    * Find out if a given category is custom/user-defined
@@ -606,13 +586,11 @@ extern "C" {
    * Check if subprotocols of the specified master protocol are just
    * informative (and not real)
    *
-   * @par     mod           = the detection module
    * @par     protoId       = the (master) protocol identifier to query
    * @return  1 = the subprotocol is informative, 0 otherwise.
    *
    */
-  u_int8_t ndpi_is_subprotocol_informative(struct ndpi_detection_module_struct *ndpi_mod,
-					   u_int16_t protoId);
+  u_int8_t ndpi_is_subprotocol_informative(u_int16_t protoId);
 
   /**
    * Set hostname-based protocol
@@ -649,6 +627,15 @@ extern "C" {
    */
   const char* ndpi_confidence_get_name(ndpi_confidence_t confidence);
 
+  /**
+   * Get FPC confidence as string
+   *
+   * @par     confidence      = the confidence value
+   * @return  the string name of the confidence result
+   *
+   */
+  const char* ndpi_fpc_confidence_get_name(ndpi_fpc_confidence_t fpc_confidence);
+
   /**
    * Set protocol category string
    *
@@ -700,8 +687,7 @@ extern "C" {
    * @return  the string name of the breed ID
    *
    */
-  char* ndpi_get_proto_breed_name(struct ndpi_detection_module_struct *ndpi_struct,
-				  ndpi_protocol_breed_t breed_id);
+  char* ndpi_get_proto_breed_name(ndpi_protocol_breed_t breed_id);
 
   /**
    * Return the ID of the protocol
@@ -763,8 +749,6 @@ extern "C" {
    */
   int ndpi_load_protocols_file(struct ndpi_detection_module_struct *ndpi_mod,
 			       const char* path);
-  int ndpi_load_protocols_file2(struct ndpi_detection_module_struct *ndpi_mod,
-			        FILE *fd);
 
   /**
    * Add an IP-address based risk mask
@@ -892,35 +876,29 @@ extern "C" {
   /**
    * Retrieve information for HTTP flows
    *
-   * @par     ndpi_mod = the detection module
    * @par     flow     = the detected flow
    * @return  the HTTP method information about the flow
    *
    */
-  ndpi_http_method ndpi_get_http_method(struct ndpi_detection_module_struct *ndpi_mod,
-					struct ndpi_flow_struct *flow);
+  ndpi_http_method ndpi_get_http_method(struct ndpi_flow_struct *flow);
 
   /**
    * Get the HTTP url
    *
-   * @par     ndpi_mod = the detection module
    * @par     flow     = the detected flow
    * @return  the HTTP method information about the flow
    *
    */
-  char* ndpi_get_http_url(struct ndpi_detection_module_struct *ndpi_mod,
-			  struct ndpi_flow_struct *flow);
+  char* ndpi_get_http_url(struct ndpi_flow_struct *flow);
 
   /**
    * Get the HTTP content-type
    *
-   * @par     ndpi_mod = the detection module
    * @par     flow     = the detected flow
    * @return  the HTTP method information about the flow
    *
    */
-  char* ndpi_get_http_content_type(struct ndpi_detection_module_struct *ndpi_mod,
-				   struct ndpi_flow_struct *flow);
+  char* ndpi_get_http_content_type(struct ndpi_flow_struct *flow);
 
   /* NDPI_PROTOCOL_TOR */
   /**
@@ -944,6 +922,7 @@ extern "C" {
    *
    */
   void* ndpi_init_automa(void);
+  void *ndpi_init_automa_domain(void);
 
   /**
    * Free Aho-Corasick automata allocated with ndpi_init_automa();
@@ -1025,11 +1004,17 @@ extern "C" {
   int ndpi_enable_loaded_categories(struct ndpi_detection_module_struct *ndpi_struct);
   void* ndpi_find_ipv4_category_userdata(struct ndpi_detection_module_struct *ndpi_str,
 					 u_int32_t saddr);
+  void* ndpi_find_ipv6_category_userdata(struct ndpi_detection_module_struct *ndpi_str,
+					 struct in6_addr *saddr);
   int ndpi_fill_ip_protocol_category(struct ndpi_detection_module_struct *ndpi_struct,
 				     struct ndpi_flow_struct *flow,
 				     u_int32_t saddr,
 				     u_int32_t daddr,
 				     ndpi_protocol *ret);
+  int ndpi_fill_ipv6_protocol_category(struct ndpi_detection_module_struct *ndpi_str,
+				       struct ndpi_flow_struct *flow,
+				       struct in6_addr *saddr, struct in6_addr *daddr,
+				      ndpi_protocol *ret);
   int ndpi_match_custom_category(struct ndpi_detection_module_struct *ndpi_struct,
 				 char *name, u_int name_len, ndpi_protocol_category_t *id);
   void ndpi_fill_protocol_category(struct ndpi_detection_module_struct *ndpi_struct,
@@ -1038,9 +1023,6 @@ extern "C" {
   int ndpi_get_custom_category_match(struct ndpi_detection_module_struct *ndpi_struct,
 				     char *name_or_ip, u_int name_len,
 				     ndpi_protocol_category_t *id);
-  int ndpi_set_detection_preferences(struct ndpi_detection_module_struct *ndpi_mod,
-				     ndpi_detection_preference pref,
-				     int value);
 
   u_int16_t ndpi_map_user_proto_id_to_ndpi_id(struct ndpi_detection_module_struct *ndpi_str,
 					      u_int16_t user_proto_id);
@@ -1059,54 +1041,23 @@ extern "C" {
   u_int ndpi_get_ndpi_num_supported_protocols(struct ndpi_detection_module_struct *ndpi_mod);
   u_int ndpi_get_ndpi_num_custom_protocols(struct ndpi_detection_module_struct *ndpi_mod);
   u_int ndpi_get_ndpi_detection_module_size(void);
-  void ndpi_set_log_level(struct ndpi_detection_module_struct *ndpi_mod, u_int l);
-  void ndpi_set_debug_bitmask(struct ndpi_detection_module_struct *ndpi_mod, NDPI_PROTOCOL_BITMASK debug_bitmask);
 
   /* Simple helper to get current time, in sec */
   u_int32_t ndpi_get_current_time(struct ndpi_flow_struct *flow);
 
   /* LRU cache */
-  struct ndpi_lru_cache* ndpi_lru_cache_init(u_int32_t num_entries, u_int32_t ttl);
+  struct ndpi_lru_cache* ndpi_lru_cache_init(u_int32_t num_entries, u_int32_t ttl, int shared);
   void ndpi_lru_free_cache(struct ndpi_lru_cache *c);
-  u_int8_t ndpi_lru_find_cache(struct ndpi_lru_cache *c, u_int32_t key,
+  u_int8_t ndpi_lru_find_cache(struct ndpi_lru_cache *c, u_int64_t key,
 			       u_int16_t *value, u_int8_t clean_key_when_found, u_int32_t now_sec);
-  void ndpi_lru_add_to_cache(struct ndpi_lru_cache *c, u_int32_t key, u_int16_t value, u_int32_t now_sec);
+  void ndpi_lru_add_to_cache(struct ndpi_lru_cache *c, u_int64_t key, u_int16_t value, u_int32_t now_sec);
   void ndpi_lru_get_stats(struct ndpi_lru_cache *c, struct ndpi_lru_cache_stats *stats);
 
-  int ndpi_get_lru_cache_stats(struct ndpi_detection_module_struct *ndpi_struct,
+  int ndpi_get_lru_cache_stats(struct ndpi_global_context *g_ctx,
+			       struct ndpi_detection_module_struct *ndpi_struct,
 			       lru_cache_type cache_type,
 			       struct ndpi_lru_cache_stats *stats);
 
-  int ndpi_get_lru_cache_size(struct ndpi_detection_module_struct *ndpi_struct,
-			      lru_cache_type cache_type,
-			      u_int32_t *num_entries);
-  int ndpi_set_lru_cache_size(struct ndpi_detection_module_struct *ndpi_struct,
-			      lru_cache_type cache_type,
-			      u_int32_t num_entries);
-
-  int ndpi_set_lru_cache_ttl(struct ndpi_detection_module_struct *ndpi_struct,
-			     lru_cache_type cache_type,
-			     u_int32_t ttl);
-  int ndpi_get_lru_cache_ttl(struct ndpi_detection_module_struct *ndpi_struct,
-			     lru_cache_type cache_type,
-			     u_int32_t *ttl);
-
-  int ndpi_set_opportunistic_tls(struct ndpi_detection_module_struct *ndpi_struct,
-				 u_int16_t proto, int value);
-  int ndpi_get_opportunistic_tls(struct ndpi_detection_module_struct *ndpi_struct,
-				 u_int16_t proto);
-
-  int ndpi_set_protocol_aggressiveness(struct ndpi_detection_module_struct *ndpi_struct,
-                                       u_int16_t proto, u_int32_t value);
-  u_int32_t ndpi_get_protocol_aggressiveness(struct ndpi_detection_module_struct *ndpi_struct,
-                                             u_int16_t proto);
-
-  int ndpi_set_monitoring_state(struct ndpi_detection_module_struct *ndpi_struct,
-                                u_int16_t proto, u_int32_t num_pkts, u_int32_t flags);
-  int ndpi_get_monitoring_state(struct ndpi_detection_module_struct *ndpi_struct,
-                                u_int16_t proto, u_int32_t *num_pkts, u_int32_t *flags);
-
-
   /**
    * Find a protocol id associated with a string automata
    *
@@ -1122,17 +1073,6 @@ extern "C" {
 				    ndpi_protocol_category_t *category,
 				    ndpi_protocol_breed_t *breed);
 
-  /**
-   * Specifies the threshold used to trigger the NDPI_TLS_CERTIFICATE_ABOUT_TO_EXPIRE
-   * flow risk that by default is set to 30 days
-   *
-   * @par    ndpi_struct  = the struct created for the protocol detection
-   * @par    days         = the number of days threshold for emitting the alert
-   *
-   */
-  void ndpi_set_tls_cert_expire_days(struct ndpi_detection_module_struct *ndpi_str,
-				     u_int8_t days);
-
   void ndpi_handle_risk_exceptions(struct ndpi_detection_module_struct *ndpi_str,
 				   struct ndpi_flow_struct *flow);
 
@@ -1152,15 +1092,14 @@ extern "C" {
   int ndpi_flowv6_flow_hash(u_int8_t l4_proto, struct ndpi_in6_addr *src_ip, struct ndpi_in6_addr *dst_ip,
 			    u_int16_t src_port, u_int16_t dst_port, u_int8_t icmp_type, u_int8_t icmp_code,
 			    u_char *hash_buf, u_int8_t hash_buf_len);
-  u_int8_t ndpi_extra_dissection_possible(struct ndpi_detection_module_struct *ndpi_struct,
-					  struct ndpi_flow_struct *flow);
+  u_int8_t ndpi_extra_dissection_possible(struct ndpi_detection_module_struct *ndpi_str,
+                                          struct ndpi_flow_struct *flow);
   u_int8_t ndpi_is_safe_ssl_cipher(u_int32_t cipher);
   const char* ndpi_cipher2str(u_int32_t cipher, char unknown_cipher[8]);
   const char* ndpi_tunnel2str(ndpi_packet_tunnel tt);
   u_int16_t ndpi_guess_host_protocol_id(struct ndpi_detection_module_struct *ndpi_struct,
 					struct ndpi_flow_struct *flow);
-  int ndpi_has_human_readeable_string(struct ndpi_detection_module_struct *ndpi_struct,
-				      char *buffer, u_int buffer_size,
+  int ndpi_has_human_readeable_string(char *buffer, u_int buffer_size,
 				      u_int8_t min_string_match_len, /* Will return 0 if no string > min_string_match_len have been found */
 				      char *outbuf, u_int outbuf_len);
   /* Return a flow info string (summarized). Does only work for DNS/HTTP/TLS/QUIC. */
@@ -1234,6 +1173,7 @@ extern "C" {
   ndpi_ptree_t* ndpi_ptree_create(void);
   int ndpi_ptree_insert(ndpi_ptree_t *tree, const ndpi_ip_addr_t *addr, u_int8_t bits, u_int64_t user_data);
   int ndpi_ptree_match_addr(ndpi_ptree_t *tree, const ndpi_ip_addr_t *addr, u_int64_t *user_data);
+  int ndpi_load_ptree_file(ndpi_ptree_t *tree, const char *path, u_int16_t protocol_id);
   void ndpi_ptree_destroy(ndpi_ptree_t *tree);
 
   /* General purpose utilities */
@@ -1330,16 +1270,6 @@ extern "C" {
    */
   int ndpi_serialize_uint32_float(ndpi_serializer *serializer, u_int32_t key, float value, const char *format /* e.f. "%.2f" */);
 
-  /**
-   * Serialize a 32-bit unsigned int key and a double value
-   * @param serializer The serializer handle
-   * @param key The field name or ID
-   * @param value The field value
-   * @param format The float value format
-   * @return 0 on success, a negative number otherwise
-   */
-  int ndpi_serialize_uint32_double(ndpi_serializer *serializer, u_int32_t key, double value, const char *format /* e.f. "%.2f" */);
-
   /**
    * Serialize a 32-bit unsigned int key and a double value
    * @param serializer The serializer handle
@@ -1753,6 +1683,7 @@ extern "C" {
 
   /* Data analysis */
   struct ndpi_analyze_struct* ndpi_alloc_data_analysis(u_int16_t _max_series_len);
+  struct ndpi_analyze_struct* ndpi_alloc_data_analysis_from_series(const u_int32_t *values, u_int16_t num_values);
   void ndpi_init_data_analysis(struct ndpi_analyze_struct *s, u_int16_t _max_series_len);
   void ndpi_free_data_analysis(struct ndpi_analyze_struct *d, u_int8_t free_pointer);
   void ndpi_reset_data_analysis(struct ndpi_analyze_struct *d);
@@ -1770,8 +1701,8 @@ extern "C" {
   float ndpi_data_stddev(struct ndpi_analyze_struct *s);
   float ndpi_data_mean(struct ndpi_analyze_struct *s);
   u_int64_t ndpi_data_last(struct ndpi_analyze_struct *s);
-  u_int32_t ndpi_data_min(struct ndpi_analyze_struct *s);
-  u_int32_t ndpi_data_max(struct ndpi_analyze_struct *s);
+  u_int64_t ndpi_data_min(struct ndpi_analyze_struct *s);
+  u_int64_t ndpi_data_max(struct ndpi_analyze_struct *s);
   float ndpi_data_ratio(u_int32_t sent, u_int32_t rcvd);
 
   /* ******************************* */
@@ -1798,6 +1729,12 @@ extern "C" {
   /* ******************************* */
 
   void ndpi_md5(const u_char *data, size_t data_len, u_char hash[16]);
+  void ndpi_sha256(const u_char *data, size_t data_len, u_int8_t sha_hash[32]);
+  
+  u_int16_t ndpi_crc16_ccit(const void* data, size_t n_bytes);
+  u_int16_t ndpi_crc16_ccit_false(const void *data, size_t n_bytes);
+  u_int16_t ndpi_crc16_xmodem(const void *data, size_t n_bytes);
+  u_int16_t ndpi_crc16_x25(const void* data, size_t n_bytes);
   u_int32_t ndpi_crc32(const void* data, size_t n_bytes);
   u_int32_t ndpi_nearest_power_of_two(u_int32_t x);
 
@@ -1831,8 +1768,7 @@ extern "C" {
 
   ndpi_risk_enum ndpi_validate_url(char *url);
 
-  u_int8_t ndpi_is_protocol_detected(struct ndpi_detection_module_struct *ndpi_str,
-				     ndpi_protocol proto);
+  u_int8_t ndpi_is_protocol_detected(ndpi_protocol proto);
   void ndpi_serialize_risk(ndpi_serializer *serializer, ndpi_risk risk);
   void ndpi_serialize_risk_score(ndpi_serializer *serializer, ndpi_risk_enum risk);
   void ndpi_serialize_confidence(ndpi_serializer *serializer, ndpi_confidence_t confidence);
@@ -1934,7 +1870,50 @@ extern "C" {
 
   /* ******************************* */
 
-  u_int32_t ndpi_quick_16_byte_hash(u_int8_t *in_16_bytes_long);
+  /*
+   * Checks if the two series are correlated using the
+   * Pearson correlation coefficient that is a value in the -1..0..+1 range
+   * where:
+   * -1 < x < 0   Negative correlation (when one changes the other series changes in opposite direction)
+   * x = 0        No correlation       (no relationship between the series)
+   * 0 < x < 1    Positive correlation (when one changes the other series changes in the same direction)
+   * (i.e. when a series increases, the other also increase and vice-versa)
+   *
+   * @par    values_a   = First series with num_values values
+   * @par    values_b   = Second series with num_values values
+   * @par    num_values = Number of series entries
+   *
+   * @return pearson correlation coefficient
+   *
+   */
+  double ndpi_pearson_correlation(u_int32_t *values_a, u_int32_t *values_b, u_int16_t num_values);
+  
+  /* ******************************* */
+  
+  /*
+   * Checks if a specified value is an outlier with respect to past values
+   * using the Z-score.
+   *
+   * @par past_valuea     = List of observed past values (past knowledge)
+   * @par num_past_values = Number of observed past values
+   * @par value_to_check  = The value to be checked with respect to past values
+   * @par threshold       = Threshold on z-score:. Typical values:
+   *                        t = 1 - The value to check should not exceed the past values
+   *                        t > 1 - The value to check has to be within (t * stddev) boundaries
+   * @par lower           - [out] Lower threshold
+   * @par upper           - [out] Upper threshold   
+   *
+   * @return true if the specified value is an outlier, false otherwise
+   *
+   */
+  
+  bool ndpi_is_outlier(u_int32_t *past_values, u_int32_t num_past_values,
+		       u_int32_t value_to_check, float threshold,
+		       float *lower, float *upper);
+
+  /* ******************************* */
+
+  u_int32_t ndpi_quick_16_byte_hash(const u_int8_t *in_16_bytes_long);
 
   /* ******************************* */
 
@@ -1952,11 +1931,9 @@ extern "C" {
    * Free the hashmap.
    *
    * @par    h            = pointer to the hash map [in, out]
-   * @par    cleanup_func = pointer to a optional callback function
-   *                        called for each element in the hashmap [in]
    *
    */
-  void ndpi_hash_free(ndpi_str_hash **h, void (*cleanup_func)(ndpi_str_hash *h));
+  void ndpi_hash_free(ndpi_str_hash **h);
 
   /**
    * Search for an entry in the hashmap.
@@ -1970,7 +1947,7 @@ extern "C" {
    * @return 0 if an entry with that key was found, 1 otherwise
    *
    */
-  int ndpi_hash_find_entry(ndpi_str_hash *h, char *key, u_int key_len, void **value);
+  int ndpi_hash_find_entry(ndpi_str_hash *h, char *key, u_int key_len, u_int16_t *value);
 
   /**
    * Add an entry to the hashmap.
@@ -1978,12 +1955,12 @@ extern "C" {
    * @par    h            = pointer to the hash map [in, out]
    * @par    key          = character string (no '\0' required) [in]
    * @par    key_len      = length of the character string @key [in]
-   * @par    value        = pointer to the value to add [in]
+   * @par    value        = value to add [in]
    *
    * @return 0 if the entry was added, 1 otherwise
    *
    */
-  int ndpi_hash_add_entry(ndpi_str_hash **h, char *key, u_int8_t key_len, void *value);
+  int ndpi_hash_add_entry(ndpi_str_hash **h, char *key, u_int8_t key_len, u_int16_t value);
 
   /* ******************************* */
 
@@ -2012,18 +1989,16 @@ extern "C" {
   */
 
   ndpi_bitmap* ndpi_bitmap_alloc(void);
-  ndpi_bitmap* ndpi_bitmap_alloc_size(u_int32_t size);
   void ndpi_bitmap_free(ndpi_bitmap* b);
   ndpi_bitmap* ndpi_bitmap_copy(ndpi_bitmap* b);
   u_int64_t ndpi_bitmap_cardinality(ndpi_bitmap* b);
   bool ndpi_bitmap_is_empty(ndpi_bitmap* b);
-  void ndpi_bitmap_set(ndpi_bitmap* b, u_int32_t value);
-  void ndpi_bitmap_unset(ndpi_bitmap* b, u_int32_t value);
-  bool ndpi_bitmap_isset(ndpi_bitmap* b, u_int32_t value);
-  void ndpi_bitmap_clear(ndpi_bitmap* b);
+  void ndpi_bitmap_set(ndpi_bitmap* b, u_int64_t value);
+  void ndpi_bitmap_unset(ndpi_bitmap* b, u_int64_t value);
+  bool ndpi_bitmap_isset(ndpi_bitmap* b, u_int64_t value);
 
   size_t ndpi_bitmap_serialize(ndpi_bitmap* b, char **buf);
-  ndpi_bitmap* ndpi_bitmap_deserialize(char *buf);
+  ndpi_bitmap* ndpi_bitmap_deserialize(char *buf, size_t buf_len);
 
   void ndpi_bitmap_and(ndpi_bitmap* a, ndpi_bitmap* b_and);
   ndpi_bitmap* ndpi_bitmap_and_alloc(ndpi_bitmap* a, ndpi_bitmap* b_and);
@@ -2035,7 +2010,7 @@ extern "C" {
 
   ndpi_bitmap_iterator* ndpi_bitmap_iterator_alloc(ndpi_bitmap* b);
   void ndpi_bitmap_iterator_free(ndpi_bitmap* b);
-  bool ndpi_bitmap_iterator_next(ndpi_bitmap_iterator* i, u_int32_t *value);
+  bool ndpi_bitmap_iterator_next(ndpi_bitmap_iterator* i, u_int64_t *value);
 
   /* ******************************* */
 
@@ -2045,16 +2020,16 @@ extern "C" {
 
     This is
     - a probabilistic datastructure !!! (i.e. be prepared to false positives)
-    - immutable (i.e. adding keys after a search (i.e. ndpi_bitmap64_isset)
+    - immutable (i.e. adding keys after a search (i.e. ndpi_bitmap64_fuse_isset)
       is not allowed
    */
 
-  ndpi_bitmap64* ndpi_bitmap64_alloc(void);
-  bool ndpi_bitmap64_set(ndpi_bitmap64 *b, u_int64_t value);
-  bool ndpi_bitmap64_compress(ndpi_bitmap64 *b);
-  bool ndpi_bitmap64_isset(ndpi_bitmap64 *b, u_int64_t value);
-  void ndpi_bitmap64_free(ndpi_bitmap64 *b);
-  u_int32_t ndpi_bitmap64_size(ndpi_bitmap64 *b);
+  ndpi_bitmap64_fuse* ndpi_bitmap64_fuse_alloc(void);
+  bool ndpi_bitmap64_fuse_set(ndpi_bitmap64_fuse *b, u_int64_t value);
+  bool ndpi_bitmap64_fuse_compress(ndpi_bitmap64_fuse *b);
+  bool ndpi_bitmap64_fuse_isset(ndpi_bitmap64_fuse *b, u_int64_t value);
+  void ndpi_bitmap64_fuse_free(ndpi_bitmap64_fuse *b);
+  u_int32_t ndpi_bitmap64_fuse_size(ndpi_bitmap64_fuse *b);
 
   /* ******************************* */
 
@@ -2100,18 +2075,20 @@ extern "C" {
   */
 
   ndpi_domain_classify* ndpi_domain_classify_alloc(void);
-  void                  ndpi_domain_classify_free(ndpi_domain_classify *s);
-  u_int32_t             ndpi_domain_classify_size(ndpi_domain_classify *s);
-  bool                  ndpi_domain_classify_add(ndpi_domain_classify *s,
-						 u_int8_t class_id, const char *domain);
-  u_int32_t             ndpi_domain_classify_add_domains(ndpi_domain_classify *s,
-							 u_int8_t class_id,
-							 char *file_path);
-  bool                  ndpi_domain_classify_finalize(ndpi_domain_classify *s);
-  bool                  ndpi_domain_classify_contains(ndpi_domain_classify *s,
-						      u_int8_t *class_id /* out */,
-						      const char *domain);
-
+  void ndpi_domain_classify_free(ndpi_domain_classify *s);
+  u_int32_t ndpi_domain_classify_size(ndpi_domain_classify *s);
+  bool ndpi_domain_classify_add(struct ndpi_detection_module_struct *ndpi_mod,
+				ndpi_domain_classify *s,
+				u_int16_t class_id, char *domain);
+  u_int32_t ndpi_domain_classify_add_domains(struct ndpi_detection_module_struct *ndpi_mod,
+					     ndpi_domain_classify *s,
+					     u_int16_t class_id,
+					     char *file_path);
+  bool ndpi_domain_classify_hostname(struct ndpi_detection_module_struct *ndpi_mod,
+				     ndpi_domain_classify *s,
+				     u_int16_t *class_id /* out */,
+				     char *hostname);
+  
   /* ******************************* */
 
   /*
@@ -2149,11 +2126,148 @@ extern "C" {
   /**
    * Get user data which was previously set with `ndpi_set_user_data()`.
    *
+   * @par ndpi_str = the struct created for the protocol detection
+   *
    * @return the user data pointer
    *
    */
   void *ndpi_get_user_data(struct ndpi_detection_module_struct *ndpi_str);
 
+  /* ******************************* */
+
+  /**
+   * Loads the domain suffixes from the specified path. You need to
+   * perform this action once
+   *
+   * @par ndpi_str = the struct created for the protocol detection
+   * @par public_suffix_list_path = path of the public_suffix_list path
+   *
+   * @return 0 = no error, -1 otherwise
+   *
+   */
+  int ndpi_load_domain_suffixes(struct ndpi_detection_module_struct *ndpi_str,
+				char *public_suffix_list_path);
+
+  /**
+   * Returns the domain suffix out of the specified hostname.
+   * The returned pointer is an offset of the original hostname.
+   * Note that you need to call ndpi_load_domain_suffixes() before
+   * calling this function.
+   *
+   * @par ndpi_str = the struct created for the protocol detection
+   * @par hostname = the hostname from which the domain name has to be extracted
+   * @par suffix_id = the id of the returned domain
+   *
+   * @return The host domain name suffic or the host itself if not found.
+   *
+   */
+  const char* ndpi_get_host_domain_suffix(struct ndpi_detection_module_struct *ndpi_str,
+					  const char *hostname,
+					  u_int16_t *suffix_id /* out */);
+
+  /**
+   * Returns the domain (including the TLS) suffix out of the specified hostname.
+   * The returned pointer is an offset of the original hostname.
+   * Note that you need to call ndpi_load_domain_suffixes() before
+   * calling this function.
+   *
+   * @par ndpi_str = the struct created for the protocol detection
+   * @par hostname = the hostname from which the domain name has to be extracted
+   *
+   * @return The host domain name or the hosti tself if not found.
+   *
+   */
+  const char* ndpi_get_host_domain(struct ndpi_detection_module_struct *ndpi_str,
+				   const char *hostname);
+
+  /* ******************************* */
+
+  ndpi_cfg_error ndpi_set_config(struct ndpi_detection_module_struct *ndpi_str,
+                                 const char *proto, const char *param, const char *value);
+  ndpi_cfg_error ndpi_set_config_u64(struct ndpi_detection_module_struct *ndpi_str,
+                                     const char *proto, const char *param, uint64_t value);
+  char *ndpi_get_config(struct ndpi_detection_module_struct *ndpi_str,
+			const char *proto, const char *param, char *buf, int buf_len);
+  char *ndpi_dump_config(struct ndpi_detection_module_struct *ndpi_str,
+			 FILE *fd);
+
+  /* ******************************* */
+
+  /* Can't call libc functions from kernel space, define some stub instead */
+
+#define ndpi_isalpha(ch) (((ch) >= 'a' && (ch) <= 'z') || ((ch) >= 'A' && (ch) <= 'Z'))
+#define ndpi_isdigit(ch) ((ch) >= '0' && (ch) <= '9')
+#define ndpi_isalnum(ch) (ndpi_isalpha(ch) != 0 || ndpi_isdigit(ch) != 0)
+#define ndpi_isspace(ch) (((ch) >= '\t' && (ch) <= '\r') || ((ch) == ' '))
+#define ndpi_isprint(ch) ((ch) >= 0x20 && (ch) <= 0x7e)
+#define ndpi_ispunct(ch) (((ch) >= '!' && (ch) <= '/') ||   \
+              ((ch) >= ':' && (ch) <= '@') ||   \
+              ((ch) >= '[' && (ch) <= '`') ||   \
+              ((ch) >= '{' && (ch) <= '~'))
+
+  /* ******************************* */
+
+  int64_t ndpi_strtonum(const char *numstr, int64_t minval, int64_t maxval, const char **errstrp, int base);
+  int ndpi_vsnprintf(char * str, size_t size, char const * format, va_list va_args);
+  int ndpi_snprintf(char * str, size_t size, char const * format, ...);
+  struct tm *ndpi_gmtime_r(const time_t *timep,
+                           struct tm *result);
+
+  /* ******************************* */
+
+  size_t ndpi_compress_str(const char * in, size_t len, char * out, size_t bufsize);
+  size_t ndpi_decompress_str(const char * in, size_t len, char * out, size_t bufsize);
+
+  /* ******************************* */
+
+  /* NOTE
+     this function works best if yout have loaded in memory domain
+     suffixes using ndpi_load_domain_suffixes()
+  */
+  u_int ndpi_encode_domain(struct ndpi_detection_module_struct *ndpi_str,
+			   char *domain, char *out, u_int out_len);
+    
+  /* ******************************* */
+
+  const char *ndpi_lru_cache_idx_to_name(lru_cache_type idx);
+
+  /**
+   * @brief Finds the first occurrence of the sequence `needle` in the array
+   * `haystack`.
+   *
+   * This function searches for the first occurrence of the sequence `needle` of
+   * length `needle_len` in the array `haystack` of length `haystack_len`. If
+   * `haystack` or `needle` is `NULL`, or `haystack_len` is less than
+   * `needle_len`, or `needle_len` is 0, the function returns `NULL`.
+   *
+   * For optimization, if `needle_len` is 1, the `memchr` function is used.
+   *
+   * @param haystack Pointer to the array in which the search is performed.
+   * @param haystack_len Length of the `haystack` array.
+   * @param needle Pointer to the array to be searched for in `haystack`.
+   * @param needle_len Length of the `needle` array.
+   * @return Pointer to the first occurrence of `needle` in `haystack` or `NULL`
+   * if `needle` is not found.
+   */
+  void* ndpi_memmem(const void* haystack, size_t haystack_len, const void* needle,
+                    size_t needle_len);
+
+  /**
+   * @brief Copies src string to dst buffer with length limit
+   *
+   * Copies the string src into dst buffer, limiting the copy length by dst_len.
+   * Handles both null-terminated and non null-terminated strings based on
+   * src_len. Ensures null-termination in dst if dst_len > 0.
+   *
+   * @param dst Destination buffer
+   * @param src Source string
+   * @param dst_len Size of dst buffer
+   * @param src_len Length of src string
+   *
+   * @return Length of src string
+   */
+  size_t ndpi_strlcpy(char* dst, const char* src, size_t dst_len, size_t src_len);
+
 #ifdef __cplusplus
 }
 #endif
