@@ -1,5 +1,5 @@
 /*
-E * ndpi_typedefs.h
+ * ndpi_typedefs.h
  *
  * Copyright (C) 2011-23 - ntop.org
  *
@@ -24,12 +24,19 @@ E * ndpi_typedefs.h
 #ifndef __NDPI_TYPEDEFS_H__
 #define __NDPI_TYPEDEFS_H__
 
+#ifndef NDPI_CFFI_PREPROCESSING
+#ifdef USE_GLOBAL_CONTEXT
+#include "ndpi_config.h"
+#define HAVE_STRUCT_TIMESPEC
+#include <pthread.h>
+#endif
+#endif
+
 #include "ndpi_define.h"
 #ifndef NDPI_CFFI_PREPROCESSING
 #include "ndpi_includes.h"
 #endif
 #include "ndpi_protocol_ids.h"
-#include "ndpi_utils.h"
 
 /* Used by both nDPI core and patricia code under third-party */
 #include "ndpi_patricia_typedefs.h"
@@ -156,11 +163,13 @@ typedef enum {
   NDPI_HTTP_OBSOLETE_SERVER,
   NDPI_PERIODIC_FLOW,          /* Set in case a flow repeats at a specific pace [used by apps on top of nDPI] */
   NDPI_MINOR_ISSUES,           /* Generic packet issues (e.g. DNS with 0 TTL) */
-  NDPI_TCP_ISSUES,             /* 50 */ /* TCP issues such as connection failed, probing or scan */
+  NDPI_TCP_ISSUES,             /* 50 */ /* TCP issues such as connection failed or scan */
   NDPI_FULLY_ENCRYPTED,        /* This (unknown) session is fully encrypted */
   NDPI_TLS_ALPN_SNI_MISMATCH,  /* Invalid ALPN/SNI combination */
   NDPI_MALWARE_HOST_CONTACTED, /* Flow client contacted a malware host */
-				 
+  NDPI_BINARY_DATA_TRANSFER,   /* Attempt to transfer something in binary format */
+  NDPI_PROBING_ATTEMPT,        /* Probing attempt (e.g. TCP connection with no data exchanged or unidirection traffic for bidirectional flows such as SSH) */
+  
   /* Leave this as last member */
   NDPI_MAX_RISK /* must be <= 63 due to (**) */
 } ndpi_risk_enum;
@@ -235,11 +244,15 @@ typedef enum {
    ndpi_leaf
 } ndpi_VISIT;
 
-/* NDPI_NODE */
-typedef struct node_t {
-  char *key;
-  struct node_t *left, *right;
-} ndpi_node;
+typedef enum {
+  NDPI_CFG_INVALID_CONTEXT = -1,
+  NDPI_CFG_NOT_FOUND = -2,
+  NDPI_CFG_INVALID_PARAM = -3,
+  NDPI_CFG_CONTEXT_ALREADY_INITIALIZED = -4,
+  NDPI_CFG_CALLBACK_ERROR = -5,
+
+  NDPI_CFG_OK = 0,
+} ndpi_cfg_error;
 
 /* NDPI_MASK_SIZE */
 typedef u_int32_t ndpi_ndpi_mask;
@@ -426,7 +439,7 @@ PACK_ON
 struct ndpi_mpls_header
 {
   /* Before using this strcut to parse an MPLS header, you will need to convert
-   * the 4-byte data to the correct endianess with ntohl(). */
+   * the 4-byte data to the correct endianness with ntohl(). */
 #if defined(__LITTLE_ENDIAN__)
   u_int32_t ttl:8, s:1, exp:3, label:20;
 #elif defined(__BIG_ENDIAN__)
@@ -585,7 +598,7 @@ struct ndpi_vxlanhdr {
 #define NDPI_GRE_STRICT      ntohs(0x0800)
 #define NDPI_GRE_REC         ntohs(0x0700)
 #define NDPI_GRE_ACK         ntohs(0x0080)
-#define NDPI_GRE_FLAGS       ntohs(0x00f8)
+#define NDPI_GRE_FLAGS       ntohs(0x0078)
 #define NDPI_GRE_VERSION     ntohs(0x0007)
 
 #define NDPI_GRE_IS_CSUM(f)		((f) & NDPI_GRE_CSUM)
@@ -635,9 +648,6 @@ struct ndpi_flow_input_info {
   unsigned char seen_flow_beginning;
 };
 
-/* Save memory limiting the key to 56 bit */
-//#define SAVE_BINARY_BITMAP_MEMORY
-
 PACK_ON
 struct ndpi_binary_bitmap_entry {
 #ifdef SAVE_BINARY_BITMAP_MEMORY
@@ -686,14 +696,25 @@ typedef enum {
 	      NDPI_HTTP_METHOD_DELETE,
 	      NDPI_HTTP_METHOD_TRACE,
 	      NDPI_HTTP_METHOD_CONNECT,
+	      NDPI_HTTP_METHOD_RPC_CONNECT,
 	      NDPI_HTTP_METHOD_RPC_IN_DATA,
 	      NDPI_HTTP_METHOD_RPC_OUT_DATA,
+	      NDPI_HTTP_METHOD_MKCOL,
+	      NDPI_HTTP_METHOD_MOVE,
+	      NDPI_HTTP_METHOD_COPY,
+	      NDPI_HTTP_METHOD_LOCK,
+	      NDPI_HTTP_METHOD_UNLOCK,
+	      NDPI_HTTP_METHOD_PROPFIND,
+	      NDPI_HTTP_METHOD_PROPPATCH,
 } ndpi_http_method;
 
 typedef enum {
   NDPI_PTREE_RISK_MASK = 0,
+  NDPI_PTREE_RISK_MASK6,
   NDPI_PTREE_RISK,
+  NDPI_PTREE_RISK6,
   NDPI_PTREE_PROTOCOLS,
+  NDPI_PTREE_PROTOCOLS6,
 
   NDPI_PTREE_MAX	/* Last one! */
 } ptree_type;
@@ -722,18 +743,22 @@ struct ndpi_automa_stats {
 typedef enum {
   NDPI_LRUCACHE_OOKLA = 0,
   NDPI_LRUCACHE_BITTORRENT,
-  NDPI_LRUCACHE_ZOOM,
   NDPI_LRUCACHE_STUN,
   NDPI_LRUCACHE_TLS_CERT,
   NDPI_LRUCACHE_MINING,
   NDPI_LRUCACHE_MSTEAMS,
-  NDPI_LRUCACHE_STUN_ZOOM,
+  NDPI_LRUCACHE_FPC_DNS, /* FPC DNS cache */
 
   NDPI_LRUCACHE_MAX	/* Last one! */
 } lru_cache_type;
 
+typedef enum {
+  NDPI_LRUCACHE_SCOPE_LOCAL = 0,
+  NDPI_LRUCACHE_SCOPE_GLOBAL,
+} lru_cache_scope;
+
 struct ndpi_lru_cache_entry {
-  u_int32_t key; /* Store the whole key to avoid ambiguities */
+  u_int64_t key; /* Store the whole key to avoid ambiguities */
   u_int32_t is_full:1, value:16, pad:15;
   u_int32_t timestamp; /* sec */
 };
@@ -746,11 +771,18 @@ struct ndpi_lru_cache_stats {
 
 struct ndpi_lru_cache {
   u_int32_t num_entries;
-  u_int32_t ttl;
+  u_int32_t ttl : 31, shared : 1;
+#ifndef NDPI_CFFI_PREPROCESSING
+#ifdef USE_GLOBAL_CONTEXT
+  pthread_mutex_t mutex;
+#endif
+#endif
   struct ndpi_lru_cache_stats stats;
   struct ndpi_lru_cache_entry *entries;
 };
 
+#define NDPI_GIVEUP_GUESS_BY_PORT	0x01
+#define NDPI_GIVEUP_GUESS_BY_IP		0x02
 
 /* Aggressiveness values */
 
@@ -759,12 +791,6 @@ struct ndpi_lru_cache {
 /* Ookla */
 #define NDPI_AGGRESSIVENESS_OOKLA_TLS			0x01 /* Enable detection over TLS (using ookla cache) */
 
-
-/* Monitoring flags */
-
-/* Stun */
-#define NDPI_MONITORING_STUN_SUBCLASSIFIED		0x01 /* Monitor STUN flows even if we have a valid sub-protocol */
-
 /* ************************************************** */
 
 struct ndpi_flow_tcp_struct {
@@ -795,15 +821,13 @@ struct ndpi_flow_tcp_struct {
 
   /* NDPI_PROTOCOL_IRC */
   u_int32_t irc_3a_counter:3;
-  u_int32_t irc_stage2:5;
-  u_int32_t irc_direction:2;
-  u_int32_t irc_0x1000_full:1;
 
   /* NDPI_PROTOCOL_USENET */
   u_int32_t usenet_stage:2;
 
   /* NDPI_PROTOCOL_HTTP */
   u_int32_t http_stage:3;
+  u_int32_t http_asymmetric_stage:2;
 
   /* NDPI_PROTOCOL_GNUTELLA */
   u_int32_t gnutella_stage:2;		       // 0 - 2
@@ -852,9 +876,6 @@ struct ndpi_flow_tcp_struct {
   /* NDPI_PROTOCOL_LOTUS_NOTES */
   u_int8_t lotus_notes_packet_id;
 
-  /* NDPI_PROTOCOL_TEAMVIEWER */
-  u_int8_t teamviewer_stage;
-
   /* NDPI_PROTOCOL_ZMQ */
   u_int8_t prev_zmq_pkt_len;
   u_char prev_zmq_pkt[10];
@@ -864,6 +885,9 @@ struct ndpi_flow_tcp_struct {
 
   /* NDPI_PROTOCOL_NEST_LOG_SINK */
   u_int8_t nest_log_sink_matches;
+
+  /* NDPI_PROTOCOL_RADMIN */
+  u_int32_t radmin_stage:1;
 };
 
 /* ************************************************** */
@@ -878,13 +902,16 @@ struct ndpi_flow_udp_struct {
   /* NDPI_PROTOCOL_XBOX */
   u_int32_t xbox_stage:1;
 
-  /* NDPI_PROTOCOL_RTP */
-  u_int32_t rtp_stage:2;
-
   /* NDPI_PROTOCOL_QUIC */
   u_int32_t quic_0rtt_found:1;
   u_int32_t quic_vn_pair:1;
 
+  /* NDPI_PROTOCOL_LOLWILDRIFT */
+  u_int32_t lolwildrift_stage:1;
+
+  /* NDPI_PROTOCOL_ZOOM */
+  u_int32_t zoom_p2p:1;
+
   /* NDPI_PROTOCOL_EPICGAMES */
   u_int32_t epicgames_stage:1;
   u_int32_t epicgames_word;
@@ -892,12 +919,6 @@ struct ndpi_flow_udp_struct {
   /* NDPI_PROTOCOL_RAKNET */
   u_int32_t raknet_custom:1;
 
-  /* NDPI_PROTOCOL_SKYPE */
-  u_int8_t skype_crc[4];
-
-  /* NDPI_PROTOCOL_TEAMVIEWER */
-  u_int8_t teamviewer_stage;
-
   /* NDPI_PROTOCOL_EAQ */
   u_int8_t eaq_pkt_id;
   u_int32_t eaq_sequence;
@@ -917,10 +938,9 @@ struct ndpi_flow_udp_struct {
   u_int8_t *quic_reasm_buf;
   u_int8_t *quic_reasm_buf_bitmap;
   u_int32_t quic_reasm_buf_last_pos;
-
-  /* NDPI_PROTOCOL_CSGO */
-  u_int8_t csgo_strid[18],csgo_state,csgo_s2;
-  u_int32_t csgo_id2;
+  /* DCID of the first Initial sent by the client */
+  u_int8_t quic_orig_dest_conn_id[20]; /* Max length is 20 on all QUIC versions */
+  u_int8_t quic_orig_dest_conn_id_len;
 
   /* NDPI_PROTOCOL_RDP */
   u_int8_t rdp_to_srv[3], rdp_from_srv[3], rdp_to_srv_pkts, rdp_from_srv_pkts;   
@@ -935,92 +955,19 @@ struct ndpi_flow_udp_struct {
   /* NDPI_PROTOCOL_TFTP */
   u_int16_t tftp_data_num;
   u_int16_t tftp_ack_num;
+
+  /* NDPI_PROTOCOL_MUMBLE */
+  u_int8_t mumble_stage:1;
+  u_int64_t mumble_ident;
 };
 
 /* ************************************************** */
 
-#define LINE_EQUALS(ndpi_int_one_line_struct, string_to_compare) \
-  ((ndpi_int_one_line_struct).len == strlen(string_to_compare) && \
-   LINE_CMP(ndpi_int_one_line_struct, string_to_compare, strlen(string_to_compare)) == 1)
-
-#define LINE_STARTS(ndpi_int_one_line_struct, string_to_compare) \
-  ((ndpi_int_one_line_struct).len >= strlen(string_to_compare) && \
-   LINE_CMP(ndpi_int_one_line_struct, string_to_compare, strlen(string_to_compare)) == 1)
-
-#define LINE_ENDS(ndpi_int_one_line_struct, string_to_compare) \
-  ((ndpi_int_one_line_struct).len >= strlen(string_to_compare) && \
-   memcmp((ndpi_int_one_line_struct).ptr + \
-          ((ndpi_int_one_line_struct).len - strlen(string_to_compare)), \
-          string_to_compare, strlen(string_to_compare)) == 0)
-
-#define LINE_CMP(ndpi_int_one_line_struct, string_to_compare, string_to_compare_length) \
-  ((ndpi_int_one_line_struct).ptr != NULL && \
-   memcmp((ndpi_int_one_line_struct).ptr, string_to_compare, string_to_compare_length) == 0)
-
-struct ndpi_int_one_line_struct {
-  const u_int8_t *ptr;
-  u_int16_t len;
-};
-
-struct ndpi_packet_struct {
-  const struct ndpi_iphdr *iph;
-  const struct ndpi_ipv6hdr *iphv6;
-  const struct ndpi_tcphdr *tcp;
-  const struct ndpi_udphdr *udp;
-  const u_int8_t *generic_l4_ptr;	/* is set only for non tcp-udp traffic */
-  const u_int8_t *payload;
-
-  u_int64_t current_time_ms;
-
-  struct ndpi_int_one_line_struct line[NDPI_MAX_PARSE_LINES_PER_PACKET];
-  /* HTTP headers */
-  struct ndpi_int_one_line_struct host_line;
-  struct ndpi_int_one_line_struct forwarded_line;
-  struct ndpi_int_one_line_struct referer_line;
-  struct ndpi_int_one_line_struct content_line;
-  struct ndpi_int_one_line_struct content_disposition_line;
-  struct ndpi_int_one_line_struct accept_line;
-  struct ndpi_int_one_line_struct authorization_line;
-  struct ndpi_int_one_line_struct user_agent_line;
-  struct ndpi_int_one_line_struct http_url_name;
-  struct ndpi_int_one_line_struct http_encoding;
-  struct ndpi_int_one_line_struct http_transfer_encoding;
-  struct ndpi_int_one_line_struct http_contentlen;
-  struct ndpi_int_one_line_struct http_cookie;
-  struct ndpi_int_one_line_struct http_origin;
-  struct ndpi_int_one_line_struct http_x_session_type;
-  struct ndpi_int_one_line_struct server_line;
-  struct ndpi_int_one_line_struct http_method;
-  struct ndpi_int_one_line_struct http_response; /* the first "word" in this pointer is the
-						    response code in the packet (200, etc) */
-  u_int8_t http_num_headers; /* number of found (valid) header lines in HTTP request or response */
-
-  u_int16_t l3_packet_len;
-  u_int16_t payload_packet_len;
-  u_int16_t parsed_lines;
-  u_int16_t empty_line_position;
-  u_int8_t tcp_retransmission;
-
-  u_int8_t packet_lines_parsed_complete:1,
-    packet_direction:1, empty_line_position_set:1, http_check_content:1, pad:4;
-};
 
+struct ndpi_global_context;
 struct ndpi_detection_module_struct;
 struct ndpi_flow_struct;
 
-struct ndpi_call_function_struct {
-  NDPI_PROTOCOL_BITMASK detection_bitmask;
-  NDPI_PROTOCOL_BITMASK excluded_protocol_bitmask;
-  void (*func) (struct ndpi_detection_module_struct *, struct ndpi_flow_struct *flow);
-  NDPI_SELECTION_BITMASK_PROTOCOL_SIZE ndpi_selection_bitmask;
-  u_int16_t ndpi_protocol_id;
-  u_int8_t detection_feature;
-};
-
-struct ndpi_subprotocol_conf_struct {
-  void (*func) (struct ndpi_detection_module_struct *, char *attr, char *value, int protocol_id);
-};
-
 typedef struct {
   u_int16_t port_low, port_high;
 } ndpi_port_range;
@@ -1035,6 +982,7 @@ typedef enum {
   NDPI_CONFIDENCE_DPI,                      /* Deep packet inspection */
   NDPI_CONFIDENCE_MATCH_BY_IP,              /* Classification obtained looking only at the IP addresses */
   NDPI_CONFIDENCE_DPI_AGGRESSIVE,           /* Aggressive DPI: it might be a false positive */
+  NDPI_CONFIDENCE_CUSTOM_RULE,              /* Matching a custom rules */
 
   /*
     IMPORTANT
@@ -1048,6 +996,24 @@ typedef enum {
   NDPI_CONFIDENCE_MAX,
 } ndpi_confidence_t;
 
+typedef enum {
+  NDPI_FPC_CONFIDENCE_UNKNOWN           = 0,    /* Unknown First Packet Classification */
+  NDPI_FPC_CONFIDENCE_IP,                       /* FPC based on IP address */
+  NDPI_FPC_CONFIDENCE_DNS,                      /* FPC based on DNS information */
+  NDPI_FPC_CONFIDENCE_DPI,                      /* FPC based on DPI information (i.e. flow classified via DPI with only one packet)*/
+
+  /*
+    IMPORTANT
+
+    Please keep in sync with
+    ndpi_fpc_confidence_get_name()
+    in ndpi_main.c
+  */
+
+  /* Last one */
+  NDPI_FPC_CONFIDENCE_MAX,
+} ndpi_fpc_confidence_t;
+
 typedef enum {
   NDPI_PROTOCOL_SAFE = 0,              /* Surely doesn't provide risks for the network. (e.g., a news site) */
   NDPI_PROTOCOL_ACCEPTABLE,            /* Probably doesn't provide risks, but could be malicious (e.g., Dropbox) */
@@ -1151,12 +1117,6 @@ typedef enum {
   NDPI_PROTOCOL_ANY_CATEGORY /* Used to handle wildcards */
 } ndpi_protocol_category_t;
 
-typedef enum {
-   ndpi_pref_direction_detect_disable = 0,
-   ndpi_pref_max_packets_to_process,
-   ndpi_pref_enable_tls_block_dissection, /* nDPI considers only those blocks past the certificate exchange */
-} ndpi_detection_preference;
-
 /* ntop extensions */
 typedef struct ndpi_proto_defaults {
   char *protoName;
@@ -1170,22 +1130,19 @@ typedef struct ndpi_proto_defaults {
   void (*func) (struct ndpi_detection_module_struct *, struct ndpi_flow_struct *flow);
 } ndpi_proto_defaults_t;
 
-typedef struct ndpi_default_ports_tree_node {
-  ndpi_proto_defaults_t *proto;
-  u_int8_t customUserProto;
-  u_int16_t default_port;
-} ndpi_default_ports_tree_node_t;
 
 typedef struct _ndpi_automa {
   void *ac_automa; /* Real type is AC_AUTOMATA_t */
   struct ndpi_automa_stats stats;
 } ndpi_automa;
 
-typedef struct ndpi_str_hash {
-  unsigned int hash;
-  void *value;
-  // u_int8_t private_data[1]; /* Avoid error C2466 and do not initiate private data with 0  */
-} ndpi_str_hash;
+typedef void ndpi_str_hash;
+
+struct ndpi_fpc_info {
+  u_int16_t master_protocol;
+  u_int16_t app_protocol;
+  ndpi_fpc_confidence_t confidence;
+};
 
 typedef struct ndpi_proto {
   /*
@@ -1205,6 +1162,7 @@ typedef struct ndpi_proto {
 
 typedef void ndpi_bitmap;
 typedef void ndpi_bitmap64;
+typedef void ndpi_bitmap64_fuse; /* probabilistic */
 typedef void ndpi_bitmap_iterator;
 typedef void ndpi_filter;
     
@@ -1217,200 +1175,9 @@ typedef struct {
 #define MAX_NUM_NDPI_DOMAIN_CLASSIFICATIONS          16
 
 typedef struct {
-  struct {
-    u_int16_t class_id;
-    ndpi_bitmap64 *domains;
-  } classes[MAX_NUM_NDPI_DOMAIN_CLASSIFICATIONS];
+  ndpi_str_hash *domains;
 } ndpi_domain_classify;
 
-#ifdef NDPI_LIB_COMPILATION
-
-/* Needed to have access to HAVE_* defines */
-#ifndef _NDPI_CONFIG_H_
-#include "ndpi_config.h"
-#define _NDPI_CONFIG_H_
-#endif
-
-/* PLEASE DO NOT REMOVE OR CHANGE THE ORDER OF WHAT IS DELIMITED BY CFFI.NDPI_MODULE_STRUCT FLAG AS IT IS USED FOR
-   PYTHON BINDINGS AUTO GENERATION */
-//CFFI.NDPI_MODULE_STRUCT
-
-typedef struct ndpi_list_struct {
-  char *value;
-  struct ndpi_list_struct *next;
-} ndpi_list;
-
-#ifdef HAVE_NBPF
-typedef struct {
-  void *tree; /* cast to nbpf_filter* */
-  u_int16_t l7_protocol;
-} nbpf_filter;
-#endif
-
-struct ndpi_detection_module_struct {
-  NDPI_PROTOCOL_BITMASK detection_bitmask;
-
-  u_int64_t current_ts;
-  u_int16_t max_packets_to_process;
-  u_int16_t num_tls_blocks_to_follow;
-  u_int8_t skip_tls_blocks_until_change_cipher:1, enable_ja3_plus:1, _notused:6;
-  u_int8_t tls_certificate_expire_in_x_days;
-  
-  void *user_data;
-  char custom_category_labels[NUM_CUSTOM_CATEGORIES][CUSTOM_CATEGORY_LABEL_LEN];
-
-  /* callback function buffer */
-  struct ndpi_call_function_struct *callback_buffer;
-  struct ndpi_call_function_struct *callback_buffer_tcp_no_payload;
-  struct ndpi_call_function_struct *callback_buffer_tcp_payload;
-  struct ndpi_call_function_struct *callback_buffer_udp;
-  struct ndpi_call_function_struct *callback_buffer_non_tcp_udp;
-  u_int32_t callback_buffer_size;
-  u_int32_t callback_buffer_size_tcp_no_payload;
-  u_int32_t callback_buffer_size_tcp_payload;
-  u_int32_t callback_buffer_size_udp;
-  u_int32_t callback_buffer_size_non_tcp_udp;
-
-  ndpi_default_ports_tree_node_t *tcpRoot, *udpRoot;
-
-  ndpi_log_level_t ndpi_log_level; /* default error */
-
-#ifdef NDPI_ENABLE_DEBUG_MESSAGES
-  /* debug callback, only set when debug is used */
-  ndpi_debug_function_ptr ndpi_debug_printf;
-  const char *ndpi_debug_print_file;
-  const char *ndpi_debug_print_function;
-  u_int32_t ndpi_debug_print_line;
-  NDPI_PROTOCOL_BITMASK debug_bitmask;
-#endif
-
-  /* misc parameters */
-  u_int32_t tcp_max_retransmission_window_size;
-
-  /* subprotocol registration handler */
-  struct ndpi_subprotocol_conf_struct subprotocol_conf[NDPI_MAX_SUPPORTED_PROTOCOLS + 1];
-
-  u_int ndpi_num_supported_protocols;
-  u_int ndpi_num_custom_protocols;
-
-  int ac_automa_finalized;
-  /* HTTP/DNS/HTTPS/QUIC host matching */
-  ndpi_automa host_automa,                     /* Used for DNS/HTTPS */
-    risky_domain_automa, tls_cert_subject_automa,
-    host_risk_mask_automa, common_alpns_automa;
-  /* IMPORTANT: please, whenever you add a new automa:
-   * update ndpi_finalize_initialization()
-   * update automa_type above
-   */
-
-  ndpi_str_hash *malicious_ja3_hashmap, *malicious_sha1_hashmap;
-
-  ndpi_list *trusted_issuer_dn;
-
-  /* Patricia trees */
-  ndpi_patricia_tree_t *ip_risk_mask_ptree;
-  ndpi_patricia_tree_t *ip_risk_ptree; 
-  ndpi_patricia_tree_t *protocols_ptree;  /* IP-based protocol detection */
-  
-  /* *** If you add a new Patricia tree, please update ptree_type above! *** */
-
-  struct {
-#ifdef USE_LEGACY_AHO_CORASICK
-    ndpi_automa hostnames, hostnames_shadow;
-#else
-    ndpi_domain_classify *sc_hostnames, *sc_hostnames_shadow;
-#endif
-    void *ipAddresses, *ipAddresses_shadow; /* Patricia */
-    u_int8_t categories_loaded;
-  } custom_categories;
-
-  u_int8_t ip_version_limit;
-
-  /* NDPI_PROTOCOL_TINC */
-  struct cache *tinc_cache;
-
-  /* NDPI_PROTOCOL_OOKLA */
-  struct ndpi_lru_cache *ookla_cache;
-  u_int32_t ookla_cache_num_entries;
-  u_int32_t ookla_cache_ttl;
-
-  /* NDPI_PROTOCOL_BITTORRENT */
-  struct ndpi_lru_cache *bittorrent_cache;
-  u_int32_t bittorrent_cache_num_entries;
-  u_int32_t bittorrent_cache_ttl;
-
-  /* NDPI_PROTOCOL_ZOOM */
-  struct ndpi_lru_cache *zoom_cache;
-  u_int32_t zoom_cache_num_entries;
-  u_int32_t zoom_cache_ttl;
-
-  /* NDPI_PROTOCOL_STUN and subprotocols */
-  struct ndpi_lru_cache *stun_cache;
-  u_int32_t stun_cache_num_entries;
-  u_int32_t stun_cache_ttl;
-  struct ndpi_lru_cache *stun_zoom_cache;
-  u_int32_t stun_zoom_cache_num_entries;
-  u_int32_t stun_zoom_cache_ttl;
-
-  /* NDPI_PROTOCOL_TLS and subprotocols */
-  struct ndpi_lru_cache *tls_cert_cache;
-  u_int32_t tls_cert_cache_num_entries;
-  int32_t tls_cert_cache_ttl;
-  
-  /* NDPI_PROTOCOL_MINING and subprotocols */
-  struct ndpi_lru_cache *mining_cache;
-  u_int32_t mining_cache_num_entries;
-  u_int32_t mining_cache_ttl;
-
-  /* NDPI_PROTOCOL_MSTEAMS */
-  struct ndpi_lru_cache *msteams_cache;
-  u_int32_t msteams_cache_num_entries;
-  u_int32_t msteams_cache_ttl;
-
-  /* *** If you add a new LRU cache, please update lru_cache_type above! *** */
-
-  int opportunistic_tls_smtp_enabled;
-  int opportunistic_tls_imap_enabled;
-  int opportunistic_tls_pop_enabled;
-  int opportunistic_tls_ftp_enabled;
-  int opportunistic_tls_stun_enabled;
-
-  u_int32_t monitoring_stun_pkts_to_process;
-  u_int32_t monitoring_stun_flags;
-
-  u_int32_t aggressiveness_ookla;
-
-  int tcp_ack_paylod_heuristic;
-  int fully_encrypted_based_on_first_pkt_heuristic;
-
-  u_int16_t ndpi_to_user_proto_id[NDPI_MAX_NUM_CUSTOM_PROTOCOLS]; /* custom protocolId mapping */
-  ndpi_proto_defaults_t proto_defaults[NDPI_MAX_SUPPORTED_PROTOCOLS+NDPI_MAX_NUM_CUSTOM_PROTOCOLS];
-
-  u_int8_t direction_detect_disable:1, /* disable internal detection of packet direction */ _pad:7;
-
-#ifdef CUSTOM_NDPI_PROTOCOLS
-  #include "../../../nDPI-custom/custom_ndpi_typedefs.h"
-#endif
-
-  /* GeoIP */
-  void *mmdb_city, *mmdb_as;
-  u_int8_t mmdb_city_loaded, mmdb_as_loaded;
-
-  /* Current packet */
-  struct ndpi_packet_struct packet;
-  const struct ndpi_flow_input_info *input_info;
-
-#ifdef HAVE_NBPF
-  u_int8_t num_nbpf_custom_proto;
-  nbpf_filter nbpf_custom_proto[MAX_NBPF_CUSTOM_PROTO];
-#endif
-
-  u_int16_t max_payload_track_len;    
-};
-
-#endif /* NDPI_LIB_COMPILATION */
-//CFFI.NDPI_MODULE_STRUCT
-
 typedef enum {
    ndpi_cipher_safe = NDPI_CIPHER_SAFE,
    ndpi_cipher_weak = NDPI_CIPHER_WEAK,
@@ -1419,6 +1186,16 @@ typedef enum {
 
 #define MAX_NUM_TLS_SIGNATURE_ALGORITHMS 16
 
+typedef struct {
+  union {
+    u_int32_t v4;
+    u_int8_t v6[16];
+  } address; /* Network-order */
+
+  u_int16_t port;
+  u_int16_t is_ipv6: 1, _pad: 15;
+} ndpi_address_port;
+  
 struct tls_heuristics {
   /*
     TLS heuristics for detecting browsers usage
@@ -1438,11 +1215,14 @@ struct ndpi_flow_struct {
   /* init parameter, internal used to set up timestamp,... */
   u_int16_t guessed_protocol_id, guessed_protocol_id_by_ip, guessed_category, guessed_header_category;
   u_int8_t l4_proto, protocol_id_already_guessed:1, fail_with_unknown:1,
-    init_finished:1, client_packet_direction:1, packet_direction:1, is_ipv6:1, first_pkt_fully_encrypted:1, _pad1: 1;
+    init_finished:1, client_packet_direction:1, packet_direction:1, is_ipv6:1, first_pkt_fully_encrypted:1, skip_entropy_check: 1;
 
   u_int16_t num_dissector_calls;
   ndpi_confidence_t confidence; /* ndpi_confidence_t */
 
+  /* First Packet Classification info */
+  struct ndpi_fpc_info fpc;
+
   /*
     if ndpi_struct->direction_detect_disable == 1
     tcp sequence number connection tracking
@@ -1526,7 +1306,8 @@ struct ndpi_flow_struct {
   } kerberos_buf;
 
   struct {
-    u_int8_t num_pkts, num_binding_requests, num_processed_pkts, maybe_dtls;
+    u_int8_t maybe_dtls : 1, is_turn : 1, pad : 6;
+    ndpi_address_port mapped_address, peer_address, relayed_address, response_origin, other_address;
   } stun;
 
   struct {
@@ -1537,7 +1318,8 @@ struct ndpi_flow_struct {
   union {
     /* the only fields useful for nDPI and ntopng */
     struct {
-      u_int8_t num_queries, num_answers, reply_code, is_query;
+      u_int8_t num_queries, num_answers, reply_code;
+      u_int8_t is_query:1, is_rsp_addr_ipv6:1, pad:6;
       u_int16_t query_type, query_class, rsp_type, edns0_udp_payload_size;
       ndpi_ip_addr_t rsp_addr; /* The first address in a DNS response packet (A and AAAA) */
       char geolocation_iata_code[4];
@@ -1567,10 +1349,10 @@ struct ndpi_flow_struct {
     struct {
       char *server_names, *advertised_alpns, *negotiated_alpn, *tls_supported_versions, *issuerDN, *subjectDN;
       u_int32_t notBefore, notAfter;
-      char ja3_client[33], ja3_server[33];
+      char ja3_client[33], ja3_server[33], ja4_client[37];
       u_int16_t server_cipher;
       u_int8_t sha1_certificate_fingerprint[20];
-      u_int8_t hello_processed:1, ch_direction:1, subprotocol_detected:1, fingerprint_set:1, _pad:4;
+      u_int8_t hello_processed:1, ch_direction:1, subprotocol_detected:1, fingerprint_set:1, webrtc:1, _pad:3;
 
 #ifdef TLS_HANDLE_SIGNATURE_ALGORITMS
       /* Under #ifdef to save memory for those who do not need them */
@@ -1682,20 +1464,15 @@ struct ndpi_flow_struct {
 
   ndpi_protocol_category_t category;
 
-  /* NDPI_PROTOCOL_REDIS */
-  u_int8_t redis_s2d_first_char, redis_d2s_first_char;
-
   /* Only packets with L5 data (ie no TCP SYN, pure ACKs, ...) */
   u_int16_t packet_counter;		      // can be 0 - 65000
   u_int16_t packet_direction_counter[2];
+  u_int8_t  packet_direction_with_payload_observed[2]; /* 0 = no packet with payload observed, 1 = at least one packet with payload observed */
 
   /* All packets even those without payload */
   u_int16_t all_packets_counter;
   u_int16_t packet_direction_complete_counter[2];      // can be 0 - 65000
 
-  /* NDPI_PROTOCOL_H323 */
-  u_int8_t h323_valid_packets;
-
   /* NDPI_PROTOCOL_BITTORRENT */
   u_int8_t bittorrent_stage;		      // can be 0 - 255
   u_int8_t bt_check_performed : 1;
@@ -1709,18 +1486,12 @@ struct ndpi_flow_struct {
   /* NDPI_PROTOCOL_SOCKS */
   u_int8_t socks5_stage:2, socks4_stage:2;      // 0 - 3
 
-  /* NDPI_PROTOCOL_EDONKEY */
-  u_int8_t edonkey_stage:2;	                // 0 - 3
-
   /* NDPI_PROTOCOL_FTP_CONTROL */
   u_int8_t ftp_control_stage:2;
 
   /* NDPI_PROTOCOL_RTMP */
   u_int8_t rtmp_stage:2;
 
-  /* NDPI_PROTOCOL_STEAM */
-  u_int16_t steam_stage:3, steam_stage1:3, steam_stage2:2, steam_stage3:2;
-
   /* NDPI_PROTOCOL_STARCRAFT */
   u_int8_t starcraft_udp_stage : 3;	// 0-7
 
@@ -1730,13 +1501,26 @@ struct ndpi_flow_struct {
   /* NDPI_PROTOCOL_OOKLA */
   u_int8_t ookla_stage : 1;
 
+  /* NDPI_PROTOCOL_TEAMVIEWER */
+  u_int8_t teamviewer_stage : 3;
+
+  /* NDPI_PROTOCOL_BFCP */
+  u_int8_t bfcp_stage:1;
+  u_int32_t bfcp_conference_id;
 
   /* NDPI_PROTOCOL_OPENVPN */
-  u_int8_t ovpn_session_id[8];
-  u_int8_t ovpn_counter;
+  u_int8_t ovpn_session_id[2][8];
 
   /* NDPI_PROTOCOL_TINC */
   u_int8_t tinc_state;
+ 
+   /* NDPI_PROTOCOL_RTCP */ 
+   u_int8_t rtcp_stage:2;
+
+   /* NDPI_PROTOCOL_RTP */
+   u_int8_t rtp_stage:2;
+   u_int8_t rtp_seq_set[2];
+   u_int16_t rtp_seq[2];
 
   /* Flow payload */
   u_int16_t flow_payload_len;
@@ -1752,11 +1536,11 @@ struct ndpi_flow_struct {
 
 #if !defined(NDPI_CFFI_PREPROCESSING) && defined(__linux__)
 #if defined(__STDC_VERSION__) && __STDC_VERSION__ >= 201112L
-_Static_assert(sizeof(((struct ndpi_flow_struct *)0)->protos) <= 216,
-               "Size of the struct member protocols increased to more than 216 bytes, "
+_Static_assert(sizeof(((struct ndpi_flow_struct *)0)->protos) <= 256,
+               "Size of the struct member protocols increased to more than 256 bytes, "
                "please check if this change is necessary.");
-_Static_assert(sizeof(struct ndpi_flow_struct) <= 968,
-               "Size of the flow struct increased to more than 968 bytes, "
+_Static_assert(sizeof(struct ndpi_flow_struct) <= 1120,
+               "Size of the flow struct increased to more than 1120 bytes, "
                "please check if this change is necessary.");
 #endif
 #endif
@@ -1771,64 +1555,17 @@ typedef struct {
   int level; /* NDPI_PROTOCOL_DEFAULT_LEVEL (0) by default */
 } ndpi_protocol_match;
 
-typedef struct {
-  char *string_to_match;
-  ndpi_protocol_category_t protocol_category;
-} ndpi_category_match;
-
-typedef struct {
-  char *string_to_match;
-  u_int16_t protocol_id;
-} ndpi_tls_cert_name_match;
-
 typedef struct {
   u_int32_t network;
   u_int8_t cidr;
   u_int16_t value;
 } ndpi_network;
 
-typedef u_int32_t ndpi_init_prefs;
-
-typedef enum {
-    ndpi_no_prefs                  = 0,
-    ndpi_dont_load_tor_list        = (1 << 0),
-    ndpi_dont_init_libgcrypt       = (1 << 1),
-    ndpi_enable_ja3_plus           = (1 << 2),
-    ndpi_dont_load_azure_list      = (1 << 3),
-    ndpi_dont_load_whatsapp_list   = (1 << 4),
-    ndpi_dont_load_amazon_aws_list = (1 << 5),
-    ndpi_dont_load_ethereum_list   = (1 << 6),
-    ndpi_dont_load_zoom_list       = (1 << 7),
-    ndpi_dont_load_cloudflare_list = (1 << 8),
-    ndpi_dont_load_microsoft_list  = (1 << 9),
-    ndpi_dont_load_google_list     = (1 << 10),
-    ndpi_dont_load_google_cloud_list = (1 << 11),
-    ndpi_dont_load_asn_lists       = (1 << 12),
-    ndpi_dont_load_icloud_private_relay_list  = (1 << 13),
-    ndpi_dont_init_risk_ptree      = (1 << 14),
-    ndpi_dont_load_cachefly_list   = (1 << 15),
-    ndpi_track_flow_payload        = (1 << 16),
-    /* In some networks, there are some anomalous TCP flows where
-       the smallest ACK packets have some kind of zero padding.
-       It looks like the IP and TCP headers in those frames wrongly consider the
-       0x00 Ethernet padding bytes as part of the TCP payload.
-       While this kind of packets is perfectly valid per-se, in some conditions
-       they might be treated by the TCP reassembler logic as (partial) overlaps,
-       deceiving the classification engine.
-       Add an heuristic to detect these packets and to ignore them, allowing
-       correct detection/classification.
-       See #1946 for other details */
-    ndpi_enable_tcp_ack_payload_heuristic = (1 << 17),
-    ndpi_dont_load_crawlers_list = (1 << 18),
-    ndpi_dont_load_protonvpn_list = (1 << 19),
-    /* Heuristic to detect fully encrypted sessions, i.e. flows where every bytes of
-       the payload is encrypted in an attempt to “look like nothing”.
-       This heuristic only analyzes the first packet of the flow.
-       See: https://www.usenix.org/system/files/sec23fall-prepub-234-wu-mingshi.pdf */
-    ndpi_disable_fully_encrypted_heuristic = (1 << 20),
-    ndpi_dont_load_protonvpn_exit_nodes_list = (1 << 21),
-    ndpi_dont_load_mullvad_list = (1 << 22),
-  } ndpi_prefs;
+typedef struct {
+  char *network;
+  u_int8_t cidr;
+  u_int16_t value;
+} ndpi_network6;
 
 typedef struct {
   u_int32_t protocol_id;
@@ -1971,7 +1708,11 @@ typedef void (*ndpi_void_fn3_t)(ndpi_patricia_node_t *node, void *data, void *us
 
 /* **************************************** */
 
-typedef struct ndpi_ptree ndpi_ptree_t;
+typedef struct ndpi_ptree
+{
+  ndpi_patricia_tree_t *v4;
+  ndpi_patricia_tree_t *v6;
+} ndpi_ptree_t;
 
 /* **************************************** */
 
